@@ -1,5 +1,5 @@
 //
-// $Id: dnstracer.c,v 1.25 2002/01/25 12:05:46 mavetju Exp $
+// $Id: dnstracer.c,v 1.31 2002/02/24 21:59:23 mavetju Exp $
 //
 
 //
@@ -57,6 +57,7 @@
 #include <time.h>
 
 #include "dnstracer_broken.h"
+#include "config.h"
 
 #define DEFAULT_RETRIES	3
 #define DEFAULT_TIMEOUT	15
@@ -121,6 +122,7 @@ struct dnsheader {
 	union {
 	    unsigned short	flags;
 	    struct bit {
+#ifdef WORDS_BIGENDIAN
 		unsigned char qr:1;
 		unsigned char opcode:4;
 		unsigned char aa:1;
@@ -129,6 +131,16 @@ struct dnsheader {
 		unsigned char ra:1;
 		unsigned char zero:3;
 		unsigned char rcode:4;
+#else
+		unsigned char rcode:4;
+		unsigned char zero:3;
+		unsigned char ra:1;
+		unsigned char rd:1;
+		unsigned char tc:1;
+		unsigned char aa:1;
+		unsigned char opcode:4;
+		unsigned char qr:1;
+#endif
 	    } bit;
 	} flags;
 	unsigned short	nquestions;
@@ -146,12 +158,13 @@ struct dnsquestion {
 
 struct dnsrr {
 	unsigned char *	domainname;
+	unsigned char *	domainname_string;
 	unsigned short	type;
 	unsigned short	class;
 	unsigned int	ttl;
 	unsigned short	datalength;
 	unsigned char *	data;
-	unsigned char * data_name;	// parsed version of data if possible
+	unsigned char * data_string;	// parsed version of data if possible
 	struct dnsrr *	next;
 };
 
@@ -246,10 +259,11 @@ char *extract_rr(struct dnssession *session,char *thisrr,struct dnsrr **rr) {
     //
     // Extract just one resource-record.
     // If the record is a ns_t_ns or ns_t_cname, cache the name 
-    // into data_name.
+    // into data_string.
     //
     domainname=getname(session,&thisrr);
     RR->domainname=strdup(domainname);
+    RR->domainname_string=strdup(printablename(domainname,1));
     RR->type=getshort(thisrr);
     RR->class=getshort(thisrr+2);
     RR->ttl=getlong(thisrr+4);
@@ -258,7 +272,7 @@ char *extract_rr(struct dnssession *session,char *thisrr,struct dnsrr **rr) {
     memcpy(RR->data,thisrr+10,RR->datalength);
 
     p=thisrr+10;
-    RR->data_name=strdup(get_resource(RR->type,session,p,1));
+    RR->data_string=strdup(get_resource(RR->type,session,p,1));
     thisrr+=10+RR->datalength;
 
     *rr=RR;
@@ -512,12 +526,11 @@ void dump_header(struct dnsheader *header) {
     printf("- Identifier:           0x%04hX\n",header->identification);
     printf("- Flags:                0x%02hX (",header->flags.flags);
 
-    if (header->flags.bit.qr) printf("Q "); else printf("R ");
+    if (header->flags.bit.qr) printf("R "); else printf("Q ");
     if (header->flags.bit.aa) printf("AA ");
     if (header->flags.bit.tc) printf("TC ");
     if (header->flags.bit.rd) printf("RD ");
     if (header->flags.bit.ra) printf("RA ");
-    if (header->flags.bit.ra) printf("RA ");
     printf(")\n");
 
     printf("- Opcode:               %hu ",header->flags.bit.opcode);
@@ -698,7 +711,7 @@ int send_data(char *server,struct dnssession *session) {
     if ((cc=sendto(session->socket,pkt,len,0,
 		(struct sockaddr *)&dest,sizeof(struct sockaddr_in)))==-1) {
 	perror("send_data/sendto");
-	exit(1);
+//	exit(1);
     }
 
     free(pkt);
@@ -798,17 +811,16 @@ struct arecord *arecords=NULL;
 void add_arecord(struct dnssession *session,struct dnsrr *rr,
 		 char *server_name,char *server_ip) {
     struct arecord *	arecord;
-    char		s[NS_MAXDNAME];
 
     arecord=(struct arecord *)calloc(1,sizeof(struct arecord));
     arecord->server_name=strdup(server_name);
     arecord->server_ip=strdup(server_ip);
     arecord->rr_name=strdup(printablename(rr->domainname,1));
 
-    if (rr->data_name==NULL)
+    if (rr->data_string==NULL)
 	arecord->rr_data=NULL;
     else
-	arecord->rr_data=strdup(rr->data_name);
+	arecord->rr_data=strdup(rr->data_string);
 
     arecord->next=arecords;
     arecords=arecord;
@@ -924,8 +936,8 @@ int is_busy(char *server) {
 // The core of this program
 //
 
-int create_session(int dnssocket,char *host,
-		    char *server_ip,char *server_name,
+int create_session(char *host,
+		    char *server_ip,char *server_name,char *server_authfor,
 		    int depth,char *prefix,int last) {
     struct dnssession * session;
     struct dnsrr *	rrauth;
@@ -943,6 +955,9 @@ int create_session(int dnssocket,char *host,
 
     printf("%s ",server_name);
 
+    if (server_authfor!=NULL)
+	printf("[%s] ",server_authfor);
+
     if (server_ip[0]==0) {
 	printf("(No IP address)");
 	return 1;
@@ -989,6 +1004,7 @@ int create_session(int dnssocket,char *host,
     //
     // Send the packet and wait for an answer.
     //
+    errorcode=0;
     for (i=0;i<global_retries;i++) {
 	send_data(server_ip,session);
 	if ((errorcode=receive_data(session))==0)
@@ -1007,13 +1023,16 @@ int create_session(int dnssocket,char *host,
     }
 
     //
-    // We have an answerRR from this server, that's enough!
-    // Also cache it for later and remove the busy flag.
+    // We have an answerRR from this server, print a message.
+    // Also cache it for later.
     //
     if (session->recv_header->nanswerRR!=0) {
 	struct dnsrr *answer;
 
-	printf("Got answer");
+	if (session->recv_header->flags.bit.aa)
+	    printf("Got authoritative answer ");
+	else
+	    printf("Got answer ");
 
 	answer=session->answer;
 	while (answer!=NULL) {
@@ -1021,6 +1040,28 @@ int create_session(int dnssocket,char *host,
 	    answer=answer->next;
 	}
 	if (global_caching) add_answer(server_ip);
+    }
+
+    //
+    // If the domainname in the authority section is the same as
+    // the one from this server, mark it as lame.
+    //
+    if (session->authority!=NULL && server_authfor!=NULL) {
+	if (session->recv_header->flags.bit.aa==0 &&
+	    strcasecmp(server_authfor,
+		       session->authority->domainname_string)==0) {
+	    printf("Lame server ");
+	    remove_busy(server_ip);
+	    return 0;
+	}
+    }
+
+
+    //
+    // If the current server has an authoritative answer, don't go
+    // further.
+    //
+    if (session->recv_header->flags.bit.aa) {
 	remove_busy(server_ip);
 	return 0;
     }
@@ -1047,13 +1088,25 @@ int create_session(int dnssocket,char *host,
 	    continue;
 	}
 
+	//
+	// If the domainname in the authority section is not a postfix
+	// of what we have, don't go there. This might happen if we are
+	// looking through cnames from different domains.
+	//
+	if (strcasecmp(rrauth->domainname_string,
+		host+strlen(host)-strlen(rrauth->domainname_string))!=0) {
+	    rrauth=rrauth->next;
+	    continue;
+	}
+
 	//
 	// Count the number of IP addresses in the additionalRR
 	// for this authorityRR
 	//
 	rradd=session->additional;
 	while (rradd!=NULL) {
-	   if (strcmp(printablename(rradd->domainname,1),rrauth->data_name)==0)
+	   if (strcmp(printablename(rradd->domainname,1),
+						    rrauth->data_string)==0)
 		found++;
 	    rradd=rradd->next;
 	}
@@ -1065,7 +1118,7 @@ int create_session(int dnssocket,char *host,
 	    //
 	    while (rradd!=NULL) {
 		if (strcmp(printablename(rradd->domainname,1),
-			   rrauth->data_name)==0)
+			   rrauth->data_string)==0)
 		    break;
 		rradd=rradd->next;
 	    }
@@ -1087,14 +1140,14 @@ int create_session(int dnssocket,char *host,
 		printf("\n");
 
 		strcpy(nextserver_name,printablename(rradd->domainname,1));
-		strcpy(nextserver_ip,rradd->data_name);
-		retval+=create_session(dnssocket,host,
-			    nextserver_ip,nextserver_name,
+		strcpy(nextserver_ip,rradd->data_string);
+		retval+=create_session(host,
+			    nextserver_ip,nextserver_name,rrauth->domainname_string,
 			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
 	    } else {
 		struct hostent *h;
 
-		strcpy(nextserver_name,rrauth->data_name);
+		strcpy(nextserver_name,rrauth->data_string);
 
 		if ((h=gethostbyname(nextserver_name))==NULL) {
 		    //
@@ -1104,8 +1157,8 @@ int create_session(int dnssocket,char *host,
 		    //
 		    printf("\n");
 		    nextserver_ip[0]=0;
-		    retval+=create_session(dnssocket,host,
-			nextserver_ip,nextserver_name,
+		    retval+=create_session(host,
+			nextserver_ip,nextserver_name,rrauth->domainname_string,
 			depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
 		} else {
 		    int count,i;
@@ -1133,8 +1186,8 @@ int create_session(int dnssocket,char *host,
 			    (unsigned char)addr_list[i][2],
 			    (unsigned char)addr_list[i][3]);
 
-			retval+=create_session(dnssocket,host,
-			    nextserver_ip,nextserver_name,
+			retval+=create_session(host,
+			    nextserver_ip,nextserver_name,rrauth->domainname_string,
 			    depth+1,s,(rrauth->next==NULL && found<=1)?1:0);
 		    }
 		}
@@ -1171,6 +1224,7 @@ int create_session(int dnssocket,char *host,
 
 void usage(void) {
     fprintf(stderr,
+	"DNSTRACER version 1.4 - (c) Edwin Groothuis - http://www.mavetju.org\n"
 	"Usage: dnstracer [options] [host]\n"
 	"\t-c: disable local caching, default enabled\n"
 	"\t-o: enable overview of received answers, default disabled\n"
@@ -1198,11 +1252,11 @@ int wsockinit(void) {
 
 
 int main(int argc,char **argv) {
-    int		dnssocket;
-    char	ch;
+    int		ch;
     char *	server_name="127.0.0.1";
     char *	server_ip="127.0.0.1";
     char	ipaddress[NS_MAXDNAME];
+    int		server_root=0;
 
 #ifndef WIN32
     //
@@ -1265,8 +1319,10 @@ int main(int argc,char **argv) {
 
 	case 's':
 	    server_name=optarg;
-	    if (strcmp(server_name,".")==0)
+	    if (strcmp(server_name,".")==0) {
 		server_name=strdup("A.ROOT-SERVERS.NET");
+		server_root=1;
+	    }
 	    break;
 
 	case 't':
@@ -1308,9 +1364,7 @@ int main(int argc,char **argv) {
 	    (unsigned char)h->h_addr_list[0][3]);
     }
 
-    create_session(dnssocket,argv[0],ipaddress,server_name,0,"",1);
-
-    close(dnssocket);
+    create_session(argv[0],ipaddress,server_name,server_root==0?NULL:".",0,"",1);
 
     printf("\n");
 
